#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <dispatch/iohandle.h>
#include <dispatch/dispatch.h>
#include "feed_s.hh"

#define MAXRAND (pow(2,31)-1)
#define MYRAND(X) (int)rint(rand()/MAXRAND*X)


class FeedHandler;
// Handler that interfaces to Primary Backup service
class PrimaryBackupService: public CORBA::ClientEventHandler
{
    private:
	Feed				*_pback;
	FeedHandler 			*_feedHandler;
	CORBA::String			_feed;

	void server_aborted(const CORBA::Object *);

    public:
	PrimaryBackupService(const char *feed_name);
	~PrimaryBackupService();

};



class FeedHandler: public Feed_impl, public IOHandler
{
    private:
	StockConsumerList	_assocList;
	PrimaryBackupService	*_svc;
	CORBA::Boolean		_timer;

	void timerExpired(long, long);
    public:
	FeedHandler(PrimaryBackupService *svc, const char *name) 
			: Feed_impl(name) {_timer = CORBA::FALSE;}
	~FeedHandler() { 
		if ( _timer )
			Dispatcher::instance().stopTimer(this);
	}

	// Feed Interface
	void	register_interest(StockConsumer *consumer, 
				const CORBA::String& stockName);
};



// ------------------ PrimaryBackup Functions----------------------------

PrimaryBackupService::PrimaryBackupService(const char *feed_name) 
			: _feed(feed_name)
{
	CORBA::BindOptions options(
				CORBA::FALSE,		// Defer binding
				CORBA::TRUE,		// Enable rebind
				CORBA::ORB::_ORB_POSTMODERN, // ORB type
				this);			// event handler
	

	// Try binding to a primary if it exists
	_pback = Feed::_bind(
				feed_name,		// object name
				NULL,			// Any host
				&options);		// Bind options
	
	if ( !_pback )	 { // No primary. 
		_feedHandler = new FeedHandler(this, feed_name);
		cout << "PrimaryBackup: Going into primary mode" << endl;
	}
	else {
		_feedHandler = NULL;
		cout << "PrimaryBackup: There is alreay a primary."
			<< endl << "Going into standby  mode" << endl;
	}
}

PrimaryBackupService::~PrimaryBackupService()
{
	if ( _feedHandler )
		_feedHandler->_release();	
	if ( _pback )
		_pback->_release();
}


// Primary must be down. Become primary
void PrimaryBackupService::server_aborted(const CORBA::Object *)
{
	_pback->_release();
	_pback = NULL;
	_feedHandler = new FeedHandler(this, (const char *)_feed);
	cout << "PrimaryBackup: Primary is down. " << endl;
	cout << "Coming up as a Primary server." << endl;
}



// Timer callback
void FeedHandler::timerExpired(long, long)
{
	_timer = CORBA::FALSE;
	CORBA::ULong num_stocks = _assocList.length();
	if ( num_stocks == 0 ) // No consumers/ no interest
		return;

	// Iterate through the assocList. Generate random prices and
	// send to consumers
	for (CORBA::ULong i= num_stocks-1; i < num_stocks; i--) {
		StockQuote quote;
		quote.stock_symbol() = _assocList[i].stock_name();
		quote.price(80.0 + rand()/MAXRAND * 40.0);
		quote.code() = MYRAND(80);
		ConsumerList& list = _assocList[i].consumers();
		CORBA::ULong num_consumers = list.length();
		for (CORBA::ULong j= num_consumers-1; j < num_consumers; j--) {
			StockConsumer *consumer = list[j];
			CORBA::Environment& env = consumer->_environment();
			consumer->receive(quote);
			if ( env.check_exception() )  // consumer must be down
				list.remove(j);		// Remove
		}
		// If no consumers, remove the entry from assocList
		if ( list.length() == 0) 
			_assocList.remove(i);
	}
	if ( _assocList.length() > 0) {
		Dispatcher::instance().startTimer(3, 0, this); // Set of timer
		_timer = CORBA::TRUE;
	}
}

void FeedHandler::register_interest(StockConsumer *consumer, 
				const CORBA::String& stockName)
{
	cout << "FeedHandler::Received a register interest in " 
			<< (const char *)stockName << endl;
		
	CORBA::ULong num_stocks = _assocList.length();

	for (CORBA::ULong i=0; i< num_stocks; i++) {
		if ( _assocList[i].stock_name() == stockName ) {
			_assocList[i].consumers().append(consumer);
			break;
		}
	}
	if ( i == num_stocks ) { // new stock
		StockConsumerAssoc *assoc = new StockConsumerAssoc;
		assoc->stock_name(stockName);
		assoc->consumers().append(consumer);
		_assocList.append(assoc);
	}
	if ( num_stocks == 0) {
		Dispatcher::instance().startTimer(3, 0, this);
		_timer = CORBA::TRUE;
	}
}





static PrimaryBackupService *svc = NULL;

void sighandler(int)
{
	if ( svc )
		delete svc;
	exit(0);
}
			
int main(int argc, char **argv)
{
	signal(SIGINT, sighandler);

	if ( argc < 2 ) {
		cout << "Usage: feedhdlr 'name' " << endl;
		return 0;
	}
	svc = new PrimaryBackupService(argv[1]);
	
	
	CORBA::BOA::impl_is_ready();
	return 0;
}
	



